==> choonvdeliver-0.1.4.tar.gz: choonvdeliver-0.1.4/src/bin/vmailmgr-dovecot-wrapper.c <==
/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-  */
/* vmailmgr-dovecot-wrapper.c
 *
 * Copyright (C) 2019  Giam Teck Choon <choon@choon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h> /* getpwnam */
#include <pwd.h> /* getpwnam */
#include <string.h>
#include <time.h>
#include <crypt.h>
#include <stdint.h>
#include <limits.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <bglibs/iobuf.h>
#include <bglibs/str.h>
#include <bglibs/dict.h>
#include "common.h"
#include "is_file.h"
#include "qmail_virtualdomains.h"
#include "vpwlookup.h"

#ifndef READFD
#define READFD 3
#endif
#ifndef ERR_BADPASSWORD
#define ERR_BADPASSWORD 1
#endif
#ifndef ERR_MISUSED
#define ERR_MISUSED 2
#endif
#ifndef ERR_INTERNAL
#define ERR_INTERNAL 111
#endif

static str domain = {0,0,0};
static str username = {0,0,0};
static str user = {0,0,0};
static str ext = {0,0,0};
static str password = {0,0,0};
static str tmpstr = {0,0,0};
static str mailhome = {0,0,0};
static str prog = {0,0,0};
static vpwentry vpw;
static char *errprefix = "Dovecot: Auth: Error: ";


static void _freemem(void)
{
	if (domain.s) str_free(&domain);
	if (username.s) str_free(&username);
	if (user.s) str_free(&user);
	if (ext.s) str_free(&ext);
	if (password.s) str_free(&password);
	if (tmpstr.s) str_free(&tmpstr);
	if (prog.s) str_free(&prog);
	if (mailhome.s) str_free(&mailhome);
}

static void _readfd(void)
{
	char buf[513];
	unsigned int buflen = 0;
	while (buflen < sizeof(buf)) {
		int r;
		do {
			r = read(READFD, buf + buflen, sizeof(buf) - buflen);
		} while ((r == -1) && (errno == EINTR));
		if (r == -1) {
			errmsg(errprefix);
			errmsgf("{Read file descriptor '}d{' error\n}", READFD);
			_freemem();
			exit(ERR_INTERNAL);
		}
		if (r == 0) break;
		buflen += r;
	}
	if (buflen >= sizeof(buf)) {
		errmsg(errprefix);
		errmsgf("{Read file descriptor '}d{' buffer too long\n}",
			READFD);
		_freemem();
		exit(ERR_INTERNAL);
	}
	char *name = buf;
	char *pass = 0;
	char *bufend = buf + sizeof(buf);
	char *ptr;
	for (ptr = buf; ptr < bufend; ptr++) {
		if (*ptr == 0) {
			++ptr;
			if (!pass) {
				pass = ptr;
				break;
			} else {
				break;
			}
		}
	}
	if (name && name[0]) {
		if (!str_copys(&user, name)) {
			errno = ENOMEM;
			errmsg(errprefix);
			errmsg("Out of memory\n");
			_freemem();
			exit(ERR_INTERNAL);
		}
	}
	if (pass && pass[0]) {
		if (!str_copys(&password, pass)) {
			errno = ENOMEM;
			errmsg(errprefix);
			errmsg("Out of memory\n");
			_freemem();
			exit(ERR_INTERNAL);
		}
	}
	close(READFD);
}

static void _escapecolon(str *s)
{
	str_truncate(&tmpstr, 0);
	unsigned int i;
	for (i=0; i<s->len; i++) {
		if (s->s[i] == ':') {
			if (!str_catb(&tmpstr, "::", 2)) {
				errno = ENOMEM;
				errmsg(errprefix);
				errmsg("Out of memory\n");
				_freemem();
				exit(ERR_INTERNAL);
			}
		} else {
			if (!str_catc(&tmpstr, s->s[i])) {
				errno = ENOMEM;
				errmsg(errprefix);
				errmsg("Out of memory\n");
				_freemem();
				exit(ERR_INTERNAL);
			}
		}
	}
	str_copy(s, &tmpstr);
}

static int verifypass(const char *plain, const char *hash)
{
	char *result;
	int ok;
	result = crypt(plain, hash);
	ok = strcmp (result, hash) == 0;
	return ok ? 1 : 0;
}

int main(int argc, char *argv[])
{
	/*
	 * To test:
	 * printf "%s\0%s\0%s\0" whatever@yourdomain password | \
	 * env QMAIL_CONTROL_DIR=$HOME /mnt/projects/c/bb3/src/bin/test \
	 * /bin/true 3<&0
	 */
	if (argc < 2) {
		/* Path to checkpassword-reply binary as the first parameter */
		errmsg(errprefix);
		errmsgf("s{: path to checkpassword-reply\n}", argv[0]);
		_freemem();
		return ERR_MISUSED;
	}
	if (!str_copys(&prog, argv[1])) {
		errmsg(errprefix);
		errmsgf("s{: Out of memory\n}", argv[0]);
		_freemem();
		return ERR_INTERNAL;
	}
	_readfd();
	if (!user.s || !user.s[0]) {
		errmsg(errprefix);
		errmsg("User not set\n");
		_freemem();
		return ERR_MISUSED;
	}
	setenv("USER", user.s, 1);
	setenv("userdb_user", user.s, 1);
	int at = str_findlast(&user, '@');
	if (at > 0) {
		if (!str_copyb(&ext, user.s, at)) {
			errno = ENOMEM;
			errmsg(errprefix);
			errmsg("Out of memory\n");
			_freemem();
			return ERR_INTERNAL;
		}
		at++;
		if (!str_copyb(&domain, user.s + at, user.len - at)) {
			errno = ENOMEM;
			errmsg(errprefix);
			errmsg("Out of memory\n");
			_freemem();
			return ERR_INTERNAL;
		}
	} else {
		errmsg(errprefix);
		errmsgf("{Invalid User '}s{'\n}", user.s);
		_freemem();
		return ERR_BADPASSWORD;
	}
	if (!ext.s || !ext.s[0]) {
		errmsg(errprefix);
		errmsg("Email User not set\n");
		_freemem();
		return ERR_BADPASSWORD;
	}
	if (!domain.s || !domain.s[0]) {
		errmsg(errprefix);
		errmsg("Domain not set\n");
		_freemem();
		return ERR_BADPASSWORD;
	}
	putenv("INSECURE_SETUID=1");
	/* Dovecot calls the script with AUTHORIZED=1 environment set when
	 * performing a userdb lookup. The script must acknowledge this by
	 * changing the environment to AUTHORIZED=2, otherwise the lookup
	 * fails. Other than that, the script works the same way as a passdb
	 * checkpassword script. If user doesn't exist, use exit code 3.
	 */
	const char *authorized = getenv("AUTHORIZED");
	if (authorized && authorized[0]) {
		//putenv("AUTHORIZED=2");
		setenv("AUTHORIZED", "2", 1);
		//errmsgf("{AUTHORIZED: }s{\n}", getenv("AUTHORIZED"));
	}
	//errmsgf("{Email Username: }s{\n}", ext.s);
	//errmsgf("{Email User: }s{\n}", user.s);
	//errmsgf("{Domain: }s{\n}", domain.s);
	if (qmail_virtualdomains(&domain, &username) == 1) {
		//errmsgf("{System Username: }s{\n}", username.s);
		struct passwd *p;
		errno = 0;
		p = getpwnam(username.s);
		if (!p) {
			errmsgf("{System Username '}s{' not found\n}",
				username.s);
			_freemem();
			return ERR_BADPASSWORD;
		}
		//setuid(p->pw_uid);
		//setgid(p->pw_gid);
		if (!str_copy2s(&tmpstr, p->pw_dir, "/passwd.cdb")) {
			errno = ENOMEM;
			errmsg(errprefix);
			errmsg("Out of memory\n");
			_freemem();
			return ERR_INTERNAL;
		}
		if (!is_file(tmpstr.s)) {
			errmsg(errprefix);
			errmsgf("s{ file not found\n}", tmpstr.s);
			_freemem();
			return ERR_INTERNAL;
		}
		if (vpwlookup(&vpw, ext.s, tmpstr.s) != 1) {
			errmsgf("{Invalid or unknown virtual user '}s{'\n}",
				ext.s);
			_freemem();
			return ERR_INTERNAL;
		}
		if (vpw.expiry < (unsigned)time(0)) {
			errmsgf("{Virtual user '}s{' has expired\n}", ext);
			_freemem();
			return ERR_INTERNAL;
		}
		if (!str_copy3s(&mailhome, p->pw_dir, "/", vpw.directory.s)) {
			errno = ENOMEM;
			errmsg(errprefix);
			errmsg("Out of memory\n");
			_freemem();
			return ERR_INTERNAL;
		}
		_escapecolon(&mailhome);
		//errmsgf("{mail_home='}s{'\n}'", mailhome.s);
		setenv("HOME", mailhome.s, 1);
		setenv("userdb_home", mailhome.s, 1);
		//str_copys(&tmpstr, "userdb_mail=maildir:");
		//str_cats(&tmpstr, mailhome.s);
		//putenv(tmpstr.s);
		str_copys(&tmpstr, "userdb_uid=");
		str_catull(&tmpstr, p->pw_uid);
		putenv(tmpstr.s);
		str_copys(&tmpstr, "userdb_gid=");
		str_catull(&tmpstr, p->pw_gid);
		putenv(tmpstr.s);
		str_copys(&tmpstr, "uid=");
		str_catull(&tmpstr, p->pw_uid);
		putenv(tmpstr.s);
		str_copys(&tmpstr, "gid=");
		str_catull(&tmpstr, p->pw_gid);
		putenv(tmpstr.s);
		uintmax_t hquota = vpw.hardquota;
		uintmax_t squota = vpw.softquota;
		uintmax_t quota = UINT_MAX;
		if (squota > hquota) {
			quota = squota;
		} else {
			quota = hquota;
		}
		if (quota == UINT_MAX) {
			// userdb_quota_rule=*:storage=0
			putenv("userdb_quota_rule=*:bytes=0");
		} else {
			str_copys(&tmpstr, "userdb_quota_rule=*:bytes=");
			str_catu(&tmpstr, quota);
			str_catc(&tmpstr, 'M');
			putenv(tmpstr.s);
		}
		setenv("EXTRA", "uid gid userdb_uid userdb_gid userdb_user userdb_home userdb_quota_rule", 1);
		//putenv("EXTRA=userdb_uid userdb_gid userdb_quota_rule userdb_home");
		if ((!authorized || !authorized[0]) && password.s && password.s[0]) {
			//errmsgf("{password='}s{'\n}'", password.s);
			//errmsgf("{password hash='}s{'\n}'", vpw.pass.s);
			if (!verifypass(password.s, vpw.pass.s)) {
				//errmsg("Password verification failed\n");
				_freemem();
				return ERR_BADPASSWORD;
			}
		}
		int cfd[2];
		CATCH_NE_ZERO(pipe(cfd), exit_cb(ERR_INTERNAL, _freemem));
		pid_t childpid;
		childpid = fork();
		switch (childpid) {
		case -1:
			errmsgf("s{ fork() failed\n}", prog.s);
			_freemem();
			exit(ERR_INTERNAL);
		case 0: /* child process */
			CATCH_NE_ZERO(setgid(p->pw_gid), exit_cb(ERR_INTERNAL, _freemem));
			CATCH_NE_ZERO(setuid(p->pw_uid), exit_cb(ERR_INTERNAL, _freemem));
			close(cfd[0]);
			CATCH_NE(dup2(cfd[1], STDOUT_FILENO), STDOUT_FILENO,
				      exit_cb(ERR_INTERNAL, _freemem));
			close(cfd[1]);
			const char *cmd[] = {prog.s, 0};
			execvp(cmd[0], (char **)cmd);
			errmsgf("{execute }s{ failed\n}", prog.s);
			_freemem();
			exit(ERR_INTERNAL);
		}
		close(cfd[1]);
		/* parent process */
		signal(SIGPIPE, SIG_IGN);
		int status;
		if (waitpid(childpid, &status, WUNTRACED) != childpid) {
			errmsg("System call to 'waitpid' failed\n");
			close(cfd[0]);
			_freemem();
			exit(ERR_INTERNAL);
		}
		if (!WIFEXITED(status)) {
			errmsgf("s{ crashed\n}", prog.s);
			close(cfd[0]);
			_freemem();
			exit(ERR_INTERNAL);
		}
		close(cfd[0]);
		_freemem();
		exit(WEXITSTATUS(status));
	} else {
		//errmsg("Username: Not Found\n");
		_freemem();
		return ERR_BADPASSWORD;
	}
	_freemem();
	//execl(argv[1], argv[1], NULL);
	return ERR_MISUSED;
}

==> choonvdeliver-0.1.5.tar.gz: choonvdeliver-0.1.5/src/bin/vmailmgr-dovecot-wrapper.c <==
/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-  */
/* vmailmgr-dovecot-wrapper.c
 *
 * Copyright (C) 2019  Giam Teck Choon <choon@choon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h> /* getpwnam */
#include <pwd.h> /* getpwnam */
#include <string.h>
#include <time.h>
#include <crypt.h>
#include <stdint.h>
#include <limits.h>
#include <sys/wait.h>
#include <signal.h>
#include <errno.h>
#include <bglibs/iobuf.h>
#include <bglibs/str.h>
#include <bglibs/dict.h>
#include "common.h"
#include "is_file.h"
#include "qmail_virtualdomains.h"
#include "vpwlookup.h"

#ifndef READFD
#define READFD 3
#endif
#ifndef ERR_BADPASSWORD
#define ERR_BADPASSWORD 1
#endif
#ifndef ERR_MISUSED
#define ERR_MISUSED 2
#endif
#ifndef ERR_INTERNAL
#define ERR_INTERNAL 111
#endif

static str domain = {0,0,0};
static str username = {0,0,0};
static str user = {0,0,0};
static str ext = {0,0,0};
static str password = {0,0,0};
static str tmpstr = {0,0,0};
static str mailhome = {0,0,0};
static str prog = {0,0,0};
static str extra = {0,0,0};
static vpwentry vpw;
static char *errprefix = "vmailmgr-dovecot-wrapper: Error: ";


static void _freemem(void)
{
	if (domain.s) str_free(&domain);
	if (username.s) str_free(&username);
	if (user.s) str_free(&user);
	if (ext.s) str_free(&ext);
	if (password.s) str_free(&password);
	if (tmpstr.s) str_free(&tmpstr);
	if (prog.s) str_free(&prog);
	if (mailhome.s) str_free(&mailhome);
	if (extra.s) str_free(&extra);
}

static void _readfd(void)
{
	char buf[513];
	unsigned int buflen = 0;
	while (buflen < sizeof(buf)) {
		int r;
		do {
			r = read(READFD, buf + buflen, sizeof(buf) - buflen);
		} while ((r == -1) && (errno == EINTR));
		if (r == -1) {
			errmsg(errprefix);
			errmsgf("{Read file descriptor '}d{' error\n}", READFD);
			_freemem();
			exit(ERR_INTERNAL);
		}
		if (r == 0) break;
		buflen += r;
	}
	if (buflen >= sizeof(buf)) {
		errmsg(errprefix);
		errmsgf("{Read file descriptor '}d{' buffer too long\n}",
			READFD);
		_freemem();
		exit(ERR_INTERNAL);
	}
	char *name = buf;
	char *pass = 0;
	char *bufend = buf + sizeof(buf);
	char *ptr;
	for (ptr = buf; ptr < bufend; ptr++) {
		if (*ptr == 0) {
			++ptr;
			if (!pass) {
				pass = ptr;
				break;
			} else {
				break;
			}
		}
	}
	if (name && name[0]) {
		if (!str_copys(&user, name)) {
			errno = ENOMEM;
			errmsg(errprefix);
			errmsg("Out of memory\n");
			_freemem();
			exit(ERR_INTERNAL);
		}
	}
	if (pass && pass[0]) {
		if (!str_copys(&password, pass)) {
			errno = ENOMEM;
			errmsg(errprefix);
			errmsg("Out of memory\n");
			_freemem();
			exit(ERR_INTERNAL);
		}
	}
	close(READFD);
}

static void _escapecolon(str *s)
{
	str_truncate(&tmpstr, 0);
	unsigned int i;
	for (i=0; i<s->len; i++) {
		if (s->s[i] == ':') {
			if (!str_catb(&tmpstr, "::", 2)) {
				errno = ENOMEM;
				errmsg(errprefix);
				errmsg("Out of memory\n");
				_freemem();
				exit(ERR_INTERNAL);
			}
		} else {
			if (!str_catc(&tmpstr, s->s[i])) {
				errno = ENOMEM;
				errmsg(errprefix);
				errmsg("Out of memory\n");
				_freemem();
				exit(ERR_INTERNAL);
			}
		}
	}
	str_copy(s, &tmpstr);
}

static int verifypass(const char *plain, const char *hash)
{
	char *result;
	int ok;
	result = crypt(plain, hash);
	ok = strcmp (result, hash) == 0;
	return ok ? 1 : 0;
}

int main(int argc, char *argv[])
{
	/*
	 * To test:
	 * printf "%s\0%s\0%s\0" whatever@yourdomain password | \
	 * env QMAIL_CONTROL_DIR=$HOME /mnt/projects/c/bb3/src/bin/test \
	 * /bin/true 3<&0
	 */
	if (argc < 2) {
		/* Path to checkpassword-reply binary as the first parameter */
		errmsg(errprefix);
		errmsgf("s{: path to checkpassword-reply\n}", argv[0]);
		_freemem();
		return ERR_MISUSED;
	}
	if (!str_copys(&prog, argv[1])) {
		errmsg(errprefix);
		errmsgf("s{: Out of memory\n}", argv[0]);
		_freemem();
		return ERR_INTERNAL;
	}
	_readfd();
	if (!user.s || !user.s[0]) {
		errmsg(errprefix);
		errmsg("User not set\n");
		_freemem();
		return ERR_MISUSED;
	}
	setenv("USER", user.s, 1);
	setenv("userdb_user", user.s, 1);
	if (!str_copys(&extra, "userdb_user")) {
		errmsg(errprefix);
		errmsgf("s{: Out of memory\n}", argv[0]);
		_freemem();
		return ERR_INTERNAL;
	}
	int at = str_findlast(&user, '@');
	if (at > 0) {
		if (!str_copyb(&ext, user.s, at)) {
			errno = ENOMEM;
			errmsg(errprefix);
			errmsg("Out of memory\n");
			_freemem();
			return ERR_INTERNAL;
		}
		at++;
		if (!str_copyb(&domain, user.s + at, user.len - at)) {
			errno = ENOMEM;
			errmsg(errprefix);
			errmsg("Out of memory\n");
			_freemem();
			return ERR_INTERNAL;
		}
	} else {
		errmsg(errprefix);
		errmsgf("{Invalid User '}s{'\n}", user.s);
		_freemem();
		return ERR_BADPASSWORD;
	}
	if (!ext.s || !ext.s[0]) {
		errmsg(errprefix);
		errmsg("Email User not set\n");
		_freemem();
		return ERR_BADPASSWORD;
	}
	if (!domain.s || !domain.s[0]) {
		errmsg(errprefix);
		errmsg("Domain not set\n");
		_freemem();
		return ERR_BADPASSWORD;
	}
	putenv("INSECURE_SETUID=1");
	/* Dovecot calls the script with AUTHORIZED=1 environment set when
	 * performing a userdb lookup. The script must acknowledge this by
	 * changing the environment to AUTHORIZED=2, otherwise the lookup
	 * fails. Other than that, the script works the same way as a passdb
	 * checkpassword script. If user doesn't exist, use exit code 3.
	 */
	const char *authorized = getenv("AUTHORIZED");
	if (authorized && authorized[0]) {
		setenv("AUTHORIZED", "2", 1);
	}
	if (qmail_virtualdomains(&domain, &username) == 1) {
		struct passwd *p;
		errno = 0;
		p = getpwnam(username.s);
		if (!p) {
			errmsgf("{System Username '}s{' not found\n}",
				username.s);
			_freemem();
			return ERR_BADPASSWORD;
		}
		if (!str_copy2s(&tmpstr, p->pw_dir, "/passwd.cdb")) {
			errno = ENOMEM;
			errmsg(errprefix);
			errmsg("Out of memory\n");
			_freemem();
			return ERR_INTERNAL;
		}
		if (!is_file(tmpstr.s)) {
			errmsg(errprefix);
			errmsgf("s{ file not found\n}", tmpstr.s);
			_freemem();
			return ERR_INTERNAL;
		}
		if (vpwlookup(&vpw, ext.s, tmpstr.s) != 1) {
			errmsgf("{Invalid or unknown virtual user '}s{'\n}",
				ext.s);
			_freemem();
			return ERR_INTERNAL;
		}
		if (vpw.expiry < (unsigned)time(0)) {
			errmsgf("{Virtual user '}s{' has expired\n}", ext);
			_freemem();
			return ERR_INTERNAL;
		}
		if (!str_copy3s(&mailhome, p->pw_dir, "/", vpw.directory.s)) {
			errno = ENOMEM;
			errmsg(errprefix);
			errmsg("Out of memory\n");
			_freemem();
			return ERR_INTERNAL;
		}
		//setenv("HOME", mailhome.s, 1);
		setenv("userdb_home", mailhome.s, 1);
		str_cats(&extra, " userdb_home");
		_escapecolon(&mailhome);
		// HOME
		str_copys(&tmpstr, "HOME=maildir:");
		str_cats(&tmpstr, mailhome.s);
		putenv(tmpstr.s);
		setenv("userdb_mail", mailhome.s, 1);
		str_cats(&extra, " userdb_mail");

		str_copys(&tmpstr, "userdb_uid=");
		str_catull(&tmpstr, p->pw_uid);
		putenv(tmpstr.s);
		str_cats(&extra, " userdb_uid");

		str_copys(&tmpstr, "userdb_gid=");
		str_catull(&tmpstr, p->pw_gid);
		putenv(tmpstr.s);
		str_cats(&extra, " userdb_gid");

		str_copys(&tmpstr, "uid=");
		str_catull(&tmpstr, p->pw_uid);
		putenv(tmpstr.s);
		str_copys(&tmpstr, "gid=");
		str_catull(&tmpstr, p->pw_gid);
		putenv(tmpstr.s);
		str_cats(&extra, " uid gid");

		uintmax_t hquota = vpw.hardquota;
		uintmax_t squota = vpw.softquota;
		uintmax_t quota = UINT_MAX;
		if (squota > hquota) {
			quota = squota;
		} else {
			quota = hquota;
		}
		if (quota == UINT_MAX) {
			// userdb_quota_rule=*:storage=0
			putenv("userdb_quota_rule=*:bytes=0");
		} else {
			str_copys(&tmpstr, "userdb_quota_rule=*:bytes=");
			str_catu(&tmpstr, quota);
			str_catc(&tmpstr, 'M');
			putenv(tmpstr.s);
		}
		str_cats(&extra, " userdb_quota_rule");
		setenv("EXTRA", extra.s, 1);
		if ((!authorized || !authorized[0]) && password.s && password.s[0]) {
			if (!verifypass(password.s, vpw.pass.s)) {
				_freemem();
				return ERR_BADPASSWORD;
			}
		}
		int cfd[2];
		CATCH_NE_ZERO(pipe(cfd), exit_cb(ERR_INTERNAL, _freemem));
		pid_t childpid;
		childpid = fork();
		switch (childpid) {
		case -1:
			errmsgf("s{ fork() failed\n}", prog.s);
			_freemem();
			exit(ERR_INTERNAL);
		case 0: /* child process */
			CATCH_NE_ZERO(setgid(p->pw_gid), exit_cb(ERR_INTERNAL, _freemem));
			CATCH_NE_ZERO(setuid(p->pw_uid), exit_cb(ERR_INTERNAL, _freemem));
			close(cfd[0]);
			CATCH_NE(dup2(cfd[1], STDOUT_FILENO), STDOUT_FILENO,
				      exit_cb(ERR_INTERNAL, _freemem));
			close(cfd[1]);
			const char *cmd[] = {prog.s, 0};
			execvp(cmd[0], (char **)cmd);
			errmsgf("{execute }s{ failed\n}", prog.s);
			_freemem();
			exit(ERR_INTERNAL);
		}
		close(cfd[1]);
		/* parent process */
		signal(SIGPIPE, SIG_IGN);
		int status;
		if (waitpid(childpid, &status, WUNTRACED) != childpid) {
			errmsg("System call to 'waitpid' failed\n");
			close(cfd[0]);
			_freemem();
			exit(ERR_INTERNAL);
		}
		if (!WIFEXITED(status)) {
			errmsgf("s{ crashed\n}", prog.s);
			close(cfd[0]);
			_freemem();
			exit(ERR_INTERNAL);
		}
		close(cfd[0]);
		_freemem();
		exit(WEXITSTATUS(status));
	} else {
		//errmsg("Username: Not Found\n");
		_freemem();
		return ERR_BADPASSWORD;
	}
	_freemem();
	return ERR_MISUSED;
}
